@@ -38,7 +38,16 @@ static inline unsigned long arch_ffz(unsigned long bitmask)
 	__asm__("bsfl %1,%0" :"=r" (bitmask) :"r" (~bitmask));
 	return bitmask;
 }
+
+static inline unsigned long long get_cpu_clock(void)
+{
+	unsigned long ret;
+
+	__asm__ __volatile__("rdtsc" : "=A" (ret));
+	return ret;
+}
+
 #define ARCH_HAVE_FFZ
-#define ARCH_HAVE_SSE
+#define ARCH_HAVE_CPU_CLOCK
 
 #endif
