@@ -45,12 +45,12 @@ static int fio_mmap_file(struct thread_data *td, struct fio_file *f,
 	}
 
 	if (!td_random(td)) {
-		if (madvise(f->mmap_ptr, length, MADV_SEQUENTIAL) < 0) {
+		if (posix_madvise(f->mmap_ptr, length, POSIX_MADV_SEQUENTIAL) < 0) {
 			td_verror(td, errno, "madvise");
 			goto err;
 		}
 	} else {
-		if (madvise(f->mmap_ptr, length, MADV_RANDOM) < 0) {
+		if (posix_madvise(f->mmap_ptr, length, POSIX_MADV_RANDOM) < 0) {
 			td_verror(td, errno, "madvise");
 			goto err;
 		}
@@ -154,17 +154,23 @@ static int fio_mmapio_queue(struct thread_data *td, struct io_u *io_u)
 			io_u->error = errno;
 			td_verror(td, io_u->error, "msync");
 		}
+	} else if (io_u->ddir == DDIR_TRIM) {
+		int ret = do_io_u_trim(td, io_u);
+
+		if (!ret)
+			td_verror(td, io_u->error, "trim");
 	}
 
+
 	/*
 	 * not really direct, but should drop the pages from the cache
 	 */
-	if (td->o.odirect && !ddir_sync(io_u->ddir)) {
+	if (td->o.odirect && ddir_rw(io_u->ddir)) {
 		if (msync(io_u->mmap_data, io_u->xfer_buflen, MS_SYNC) < 0) {
 			io_u->error = errno;
 			td_verror(td, io_u->error, "msync");
 		}
-		if (madvise(io_u->mmap_data, io_u->xfer_buflen,  MADV_DONTNEED) < 0) {
+		if (posix_madvise(io_u->mmap_data, io_u->xfer_buflen, POSIX_MADV_DONTNEED) < 0) {
 			io_u->error = errno;
 			td_verror(td, io_u->error, "madvise");
 		}
@@ -175,8 +181,16 @@ static int fio_mmapio_queue(struct thread_data *td, struct io_u *io_u)
 
 static int fio_mmapio_init(struct thread_data *td)
 {
+	struct thread_options *o = &td->o;
 	unsigned long shift, mask;
 
+	if ((td->o.rw_min_bs & page_mask) &&
+	    (o->odirect || o->fsync_blocks || o->fdatasync_blocks)) {
+		log_err("fio: mmap options dictate a minimum block size of "
+			"%lu bytes\n", page_size);
+		return 1;
+	}
+
 	mmap_map_size = MMAP_TOTAL_SZ / td->o.nr_files;
 	mask = mmap_map_size;
 	shift = 0;
