@@ -24,10 +24,12 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
+#include <limits.h>
 #include <signal.h>
 #include <time.h>
 #include <locale.h>
 #include <assert.h>
+#include <time.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <sys/ipc.h>
@@ -38,9 +40,12 @@
 #include "hash.h"
 #include "smalloc.h"
 #include "verify.h"
+#include "trim.h"
 #include "diskutil.h"
 #include "cgroup.h"
 #include "profile.h"
+#include "lib/rand.h"
+#include "memalign.h"
 
 unsigned long page_mask;
 unsigned long page_size;
@@ -60,8 +65,8 @@ static struct fio_mutex *startup_mutex;
 static struct fio_mutex *writeout_mutex;
 static volatile int fio_abort;
 static int exit_value;
-static struct itimerval itimer;
 static pthread_t gtod_thread;
+static pthread_t disk_util_thread;
 static struct flist_head *cgroup_list;
 static char *cgroup_mnt;
 
@@ -97,91 +102,87 @@ static void terminate_threads(int group_id)
 			/*
 			 * if the thread is running, just let it exit
 			 */
-			if (td->runstate < TD_RUNNING)
-				kill(td->pid, SIGQUIT);
+			if (!td->pid)
+				continue;
+			else if (td->runstate < TD_RAMP)
+				kill(td->pid, SIGTERM);
 			else {
 				struct ioengine_ops *ops = td->io_ops;
 
-				if (ops && (ops->flags & FIO_SIGQUIT))
-					kill(td->pid, SIGQUIT);
+				if (ops && (ops->flags & FIO_SIGTERM))
+					kill(td->pid, SIGTERM);
 			}
 		}
 	}
 }
 
-static void status_timer_arm(void)
-{
-	itimer.it_value.tv_sec = 0;
-	itimer.it_value.tv_usec = DISK_UTIL_MSEC * 1000;
-	setitimer(ITIMER_REAL, &itimer, NULL);
-}
-
-static void sig_alrm(int fio_unused sig)
-{
-	if (threads) {
-		update_io_ticks();
-		print_thread_status();
-		status_timer_arm();
-	}
-}
-
-/*
- * Happens on thread runs with ctrl-c, ignore our own SIGQUIT
- */
-static void sig_quit(int sig)
-{
-}
-
 static void sig_int(int sig)
 {
 	if (threads) {
 		log_info("\nfio: terminating on signal %d\n", sig);
 		fflush(stdout);
+		exit_value = 128;
 		terminate_threads(TERMINATE_ALL);
 	}
 }
 
-static void sig_ill(int fio_unused sig)
+static void *disk_thread_main(void *data)
 {
-	if (!threads)
-		return;
+	fio_mutex_up(startup_mutex);
 
-	log_err("fio: illegal instruction. your cpu does not support "
-		"the sse4.2 instruction for crc32c\n");
-	terminate_threads(TERMINATE_ALL);
-	exit(4);
+	while (threads) {
+		usleep(DISK_UTIL_MSEC * 1000);
+		if (!threads)
+			break;
+		update_io_ticks();
+		print_thread_status();
+	}
+
+	return NULL;
+}
+
+static int create_disk_util_thread(void)
+{
+	int ret;
+
+	ret = pthread_create(&disk_util_thread, NULL, disk_thread_main, NULL);
+	if (ret) {
+		log_err("Can't create disk util thread: %s\n", strerror(ret));
+		return 1;
+	}
+
+	ret = pthread_detach(disk_util_thread);
+	if (ret) {
+		log_err("Can't detatch disk util thread: %s\n", strerror(ret));
+		return 1;
+	}
+
+	dprint(FD_MUTEX, "wait on startup_mutex\n");
+	fio_mutex_down(startup_mutex);
+	dprint(FD_MUTEX, "done waiting on startup_mutex\n");
+	return 0;
 }
 
 static void set_sig_handlers(void)
 {
 	struct sigaction act;
 
-	memset(&act, 0, sizeof(act));
-	act.sa_handler = sig_alrm;
-	act.sa_flags = SA_RESTART;
-	sigaction(SIGALRM, &act, NULL);
-
 	memset(&act, 0, sizeof(act));
 	act.sa_handler = sig_int;
 	act.sa_flags = SA_RESTART;
 	sigaction(SIGINT, &act, NULL);
 
 	memset(&act, 0, sizeof(act));
-	act.sa_handler = sig_ill;
+	act.sa_handler = sig_int;
 	act.sa_flags = SA_RESTART;
-	sigaction(SIGILL, &act, NULL);
-
-	memset(&act, 0, sizeof(act));
-	act.sa_handler = sig_quit;
-	act.sa_flags = SA_RESTART;
-	sigaction(SIGQUIT, &act, NULL);
+	sigaction(SIGTERM, &act, NULL);
 }
 
 /*
  * Check if we are above the minimum rate given.
  */
 static int __check_min_rate(struct thread_data *td, struct timeval *now,
-			    enum td_ddir ddir)
+			    enum fio_ddir ddir)
 {
 	unsigned long long bytes = 0;
 	unsigned long iops = 0;
@@ -191,6 +192,8 @@ static int __check_min_rate(struct thread_data *td, struct timeval *now,
 	unsigned int rate_iops = 0;
 	unsigned int rate_iops_min = 0;
 
+	assert(ddir_rw(ddir));
+
 	if (!td->o.ratemin[ddir] && !td->o.rate_iops_min[ddir])
 		return 0;
 
@@ -371,10 +374,15 @@ requeue:
 	return 0;
 }
 
+static inline void __update_tv_cache(struct thread_data *td)
+{
+	fio_gettime(&td->tv_cache, NULL);
+}
+
 static inline void update_tv_cache(struct thread_data *td)
 {
 	if ((++td->tv_cache_nr & td->tv_cache_mask) == td->tv_cache_mask)
-		fio_gettime(&td->tv_cache, NULL);
+		__update_tv_cache(td);
 }
 
 static int break_on_this_error(struct thread_data *td, int *retptr)
@@ -460,8 +468,11 @@ static void do_verify(struct thread_data *td)
 		update_tv_cache(td);
 
 		if (runtime_exceeded(td, &td->tv_cache)) {
-			td->terminate = 1;
-			break;
+			__update_tv_cache(td);
+			if (runtime_exceeded(td, &td->tv_cache)) {
+				td->terminate = 1;
+				break;
+			}
 		}
 
 		io_u = __get_io_u(td);
@@ -491,7 +502,6 @@ static void do_verify(struct thread_data *td)
 				clear_io_u(td, io_u);
 			} else if (io_u->resid) {
 				int bytes = io_u->xfer_buflen - io_u->resid;
-				struct fio_file *f = io_u->file;
 
 				/*
 				 * zero read, fail
@@ -506,8 +516,10 @@ static void do_verify(struct thread_data *td)
 				io_u->xfer_buf += bytes;
 				io_u->offset += bytes;
 
-				td->ts.short_io_u[io_u->ddir]++;
+				if (ddir_rw(io_u->ddir))
+					td->ts.short_io_u[io_u->ddir]++;
 
+				f = io_u->file;
 				if (io_u->offset == f->real_file_size)
 					goto sync_done;
 
@@ -538,9 +550,10 @@ sync_done:
 
 		/*
 		 * if we can queue more, do so. but check if there are
-		 * completed io_u's first.
+		 * completed io_u's first. Note that we can get BUSY even
+		 * without IO queued, if the system is resource starved.
 		 */
-		full = queue_full(td) || ret == FIO_Q_BUSY;
+		full = queue_full(td) || (ret == FIO_Q_BUSY && td->cur_depth);
 		if (full || !td->o.iodepth_batch_complete) {
 			min_events = min(td->o.iodepth_batch_complete,
 					 td->cur_depth);
@@ -591,6 +604,7 @@ static void do_io(struct thread_data *td)
 		td_set_runstate(td, TD_RUNNING);
 
 	while ( (td->o.read_iolog_file && !flist_empty(&td->io_log_list)) ||
+		(!flist_empty(&td->trim_list)) ||
 	        ((td->this_io_bytes[0] + td->this_io_bytes[1]) < td->o.size) ) {
 		struct timeval comp_time;
 		unsigned long bytes_done[2] = { 0, 0 };
@@ -604,8 +618,11 @@ static void do_io(struct thread_data *td)
 		update_tv_cache(td);
 
 		if (runtime_exceeded(td, &td->tv_cache)) {
-			td->terminate = 1;
-			break;
+			__update_tv_cache(td);
+			if (runtime_exceeded(td, &td->tv_cache)) {
+				td->terminate = 1;
+				break;
+			}
 		}
 
 		io_u = get_io_u(td);
@@ -651,7 +668,8 @@ static void do_io(struct thread_data *td)
 				io_u->xfer_buf += bytes;
 				io_u->offset += bytes;
 
-				td->ts.short_io_u[io_u->ddir]++;
+				if (ddir_rw(io_u->ddir))
+					td->ts.short_io_u[io_u->ddir]++;
 
 				if (io_u->offset == f->real_file_size)
 					goto sync_done;
@@ -693,9 +711,11 @@ sync_done:
 			break;
 
 		/*
-		 * See if we need to complete some commands
+		 * See if we need to complete some commands. Note that we
+		 * can get BUSY even without IO queued, if the system is
+		 * resource starved.
 		 */
-		full = queue_full(td) || ret == FIO_Q_BUSY;
+		full = queue_full(td) || (ret == FIO_Q_BUSY && td->cur_depth);
 		if (full || !td->o.iodepth_batch_complete) {
 			min_evts = min(td->o.iodepth_batch_complete,
 					td->cur_depth);
@@ -745,6 +765,9 @@ sync_done:
 		}
 	}
 
+	if (td->trim_entries)
+		printf("trim entries %ld\n", td->trim_entries);
+
 	if (td->o.fill_device && td->error == ENOSPC) {
 		td->error = 0;
 		td->terminate = 1;
@@ -753,8 +776,11 @@ sync_done:
 		struct fio_file *f;
 
 		i = td->cur_depth;
-		if (i)
+		if (i) {
 			ret = io_u_queued_complete(td, i, NULL);
+			if (td->o.fill_device && td->error == ENOSPC)
+				td->error = 0;
+		}
 
 		if (should_fsync(td) && td->o.end_fsync) {
 			td_set_runstate(td, TD_FSYNCING);
@@ -784,7 +810,7 @@ static void cleanup_io_u(struct thread_data *td)
 		io_u = flist_entry(entry, struct io_u, list);
 
 		flist_del(&io_u->list);
-		free(io_u);
+		fio_memfree(io_u, sizeof(*io_u));
 	}
 
 	free_io_mem(td);
@@ -817,7 +843,8 @@ static int init_io_u(struct thread_data *td)
 	if (allocate_io_mem(td))
 		return 1;
 
-	if (td->o.odirect || td->o.mem_align)
+	if (td->o.odirect || td->o.mem_align ||
+	    (td->io_ops->flags & FIO_RAWIO))
 		p = PAGE_ALIGN(td->orig_buffer) + td->o.mem_align;
 	else
 		p = td->orig_buffer;
@@ -830,8 +857,9 @@ static int init_io_u(struct thread_data *td)
 		if (td->terminate)
 			return 1;
 
-		if (posix_memalign(&ptr, cl_align, sizeof(*io_u))) {
-			log_err("fio: posix_memalign=%s\n", strerror(errno));
+		ptr = fio_memalign(cl_align, sizeof(*io_u));
+		if (!ptr) {
+			log_err("fio: unable to allocate aligned memory\n");
 			break;
 		}
 
@@ -846,6 +874,13 @@ static int init_io_u(struct thread_data *td)
 
 			if (td_write(td) && !td->o.refill_buffers)
 				io_u_fill_buffer(td, io_u, max_bs);
+			else if (td_write(td) && td->o.verify_pattern_bytes) {
+				/*
+				 * Fill the buffer with the pattern if we are
+				 * going to be doing writes.
+				 */
+				fill_pattern(td, io_u->buf, max_bs, io_u, 0, 0);
+			}
 		}
 
 		io_u->index = i;
@@ -948,11 +983,6 @@ static void reset_io_counters(struct thread_data *td)
 	 */
 	if (td->o.time_based || td->o.loops)
 		td->nr_done_files = 0;
-
-	/*
-	 * Set the same seed to get repeatable runs
-	 */
-	td_fill_rand_seeds(td);
 }
 
 void reset_all_stats(struct thread_data *td)
@@ -968,8 +998,10 @@ void reset_all_stats(struct thread_data *td)
 		td->io_issues[i] = 0;
 		td->ts.total_io_u[i] = 0;
 	}
-	
+
 	fio_gettime(&tv, NULL);
+	td->ts.runtime[0] = 0;
+	td->ts.runtime[1] = 0;
 	memcpy(&td->epoch, &tv, sizeof(tv));
 	memcpy(&td->start, &tv, sizeof(tv));
 }
@@ -984,6 +1016,11 @@ static void clear_io_state(struct thread_data *td)
 	close_files(td);
 	for_each_file(td, f, i)
 		fio_file_clear_done(f);
+
+	/*
+	 * Set the same seed to get repeatable runs
+	 */
+	td_fill_rand_seeds(td);
 }
 
 static int exec_string(const char *string)
@@ -1008,15 +1045,16 @@ static int exec_string(const char *string)
  */
 static void *thread_main(void *data)
 {
-	unsigned long long runtime[2], elapsed;
+	unsigned long long elapsed;
 	struct thread_data *td = data;
 	pthread_condattr_t attr;
 	int clear_state;
 
-	if (!td->o.use_thread)
+	if (!td->o.use_thread) {
 		setsid();
-
-	td->pid = getpid();
+		td->pid = getpid();
+	} else
+		td->pid = gettid();
 
 	dprint(FD_PROCESS, "jobs pid=%d started\n", (int) td->pid);
 
@@ -1026,6 +1064,7 @@ static void *thread_main(void *data)
 	INIT_FLIST_HEAD(&td->io_log_list);
 	INIT_FLIST_HEAD(&td->io_hist_list);
 	INIT_FLIST_HEAD(&td->verify_list);
+	INIT_FLIST_HEAD(&td->trim_list);
 	pthread_mutex_init(&td->io_u_lock, NULL);
 	td->io_hist_tree = RB_ROOT;
 
@@ -1046,12 +1085,32 @@ static void *thread_main(void *data)
 	 */
 	fio_mutex_remove(td->mutex);
 
+	/*
+	 * A new gid requires privilege, so we need to do this before setting
+	 * the uid.
+	 */
+	if (td->o.gid != -1U && setgid(td->o.gid)) {
+		td_verror(td, errno, "setgid");
+		goto err;
+	}
 	if (td->o.uid != -1U && setuid(td->o.uid)) {
 		td_verror(td, errno, "setuid");
 		goto err;
 	}
-	if (td->o.gid != -1U && setgid(td->o.gid)) {
-		td_verror(td, errno, "setgid");
+
+	/*
+	 * If we have a gettimeofday() thread, make sure we exclude that
+	 * thread from this job
+	 */
+	if (td->o.gtod_cpu)
+		fio_cpu_clear(&td->o.cpumask, td->o.gtod_cpu);
+
+	/*
+	 * Set affinity first, in case it has an impact on the memory
+	 * allocations.
+	 */
+	if (td->o.cpumask_set && fio_setaffinity(td->pid, td->o.cpumask) == -1) {
+		td_verror(td, errno, "cpu_set_affinity");
 		goto err;
 	}
 
@@ -1068,23 +1127,6 @@ static void *thread_main(void *data)
 	if (td->o.verify_async && verify_async_init(td))
 		goto err;
 
-	if (td->o.cpumask_set && fio_setaffinity(td->pid, td->o.cpumask) == -1) {
-		td_verror(td, errno, "cpu_set_affinity");
-		goto err;
-	}
-
-	/*
-	 * If we have a gettimeofday() thread, make sure we exclude that
-	 * thread from this job
-	 */
-	if (td->o.gtod_cpu) {
-		fio_cpu_clear(&td->o.cpumask, td->o.gtod_cpu);
-		if (fio_setaffinity(td->pid, td->o.cpumask) == -1) {
-			td_verror(td, errno, "cpu_set_affinity");
-			goto err;
-		}
-	}
-
 	if (td->ioprio_set) {
 		if (ioprio_set(IOPRIO_WHO_PROCESS, 0, td->ioprio) == -1) {
 			td_verror(td, errno, "ioprio_set");
@@ -1125,7 +1167,6 @@ static void *thread_main(void *data)
 	fio_gettime(&td->epoch, NULL);
 	getrusage(RUSAGE_SELF, &td->ts.ru_start);
 
-	runtime[0] = runtime[1] = 0;
 	clear_state = 0;
 	while (keep_running(td)) {
 		fio_gettime(&td->start, NULL);
@@ -1150,11 +1191,11 @@ static void *thread_main(void *data)
 
 		if (td_read(td) && td->io_bytes[DDIR_READ]) {
 			elapsed = utime_since_now(&td->start);
-			runtime[DDIR_READ] += elapsed;
+			td->ts.runtime[DDIR_READ] += elapsed;
 		}
 		if (td_write(td) && td->io_bytes[DDIR_WRITE]) {
 			elapsed = utime_since_now(&td->start);
-			runtime[DDIR_WRITE] += elapsed;
+			td->ts.runtime[DDIR_WRITE] += elapsed;
 		}
 
 		if (td->error || td->terminate)
@@ -1171,15 +1212,15 @@ static void *thread_main(void *data)
 
 		do_verify(td);
 
-		runtime[DDIR_READ] += utime_since_now(&td->start);
+		td->ts.runtime[DDIR_READ] += utime_since_now(&td->start);
 
 		if (td->error || td->terminate)
 			break;
 	}
 
 	update_rusage_stat(td);
-	td->ts.runtime[0] = (runtime[0] + 999) / 1000;
-	td->ts.runtime[1] = (runtime[1] + 999) / 1000;
+	td->ts.runtime[0] = (td->ts.runtime[0] + 999) / 1000;
+	td->ts.runtime[1] = (td->ts.runtime[1] + 999) / 1000;
 	td->ts.total_run_time = mtime_since_now(&td->epoch);
 	td->ts.io_bytes[0] = td->io_bytes[0];
 	td->ts.io_bytes[1] = td->io_bytes[1];
@@ -1192,6 +1233,13 @@ static void *thread_main(void *data)
 		} else
 			finish_log(td, td->ts.bw_log, "bw");
 	}
+	if (td->ts.lat_log) {
+		if (td->o.lat_log_file) {
+			finish_log_named(td, td->ts.lat_log,
+						td->o.lat_log_file, "lat");
+		} else
+			finish_log(td, td->ts.lat_log, "lat");
+	}
 	if (td->ts.slat_log) {
 		if (td->o.lat_log_file) {
 			finish_log_named(td, td->ts.slat_log,
@@ -1252,6 +1300,7 @@ static int fork_main(int shmid, int offset)
 	struct thread_data *td;
 	void *data, *ret;
 
+#ifndef __hpux
 	data = shmat(shmid, NULL, 0);
 	if (data == (void *) -1) {
 		int __err = errno;
@@ -1259,6 +1308,12 @@ static int fork_main(int shmid, int offset)
 		perror("shmat");
 		return __err;
 	}
+#else
+	/*
+	 * HP-UX inherits shm mappings?
+	 */
+	data = threads;
+#endif
 
 	td = data + offset * sizeof(struct thread_data);
 	ret = thread_main(td);
@@ -1324,7 +1379,7 @@ static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
 			if (WIFSIGNALED(status)) {
 				int sig = WTERMSIG(status);
 
-				if (sig != SIGQUIT)
+				if (sig != SIGTERM)
 					log_err("fio: pid=%d, got signal=%d\n",
 							(int) td->pid, sig);
 				td_set_runstate(td, TD_REAPED);
@@ -1381,9 +1436,13 @@ static void *gtod_thread_main(void *data)
 
 static int fio_start_gtod_thread(void)
 {
+	pthread_attr_t attr;
 	int ret;
 
-	ret = pthread_create(&gtod_thread, NULL, gtod_thread_main, NULL);
+	pthread_attr_init(&attr);
+	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN);
+	ret = pthread_create(&gtod_thread, &attr, gtod_thread_main, NULL);
+	pthread_attr_destroy(&attr);
 	if (ret) {
 		log_err("Can't create gtod thread: %s\n", strerror(ret));
 		return 1;
@@ -1460,14 +1519,14 @@ static void run_threads(void)
 			todo--;
 		} else {
 			struct fio_file *f;
-			unsigned int i;
+			unsigned int j;
 
 			/*
 			 * for sharing to work, each job must always open
 			 * its own files. so close them, if we opened them
 			 * for creation
 			 */
-			for_each_file(td, f, i) {
+			for_each_file(td, f, j) {
 				if (fio_file_open(f))
 					td_io_close_file(td, f);
 			}
@@ -1479,7 +1538,7 @@ static void run_threads(void)
 	set_genesis_time();
 
 	while (todo) {
-		struct thread_data *map[MAX_JOBS];
+		struct thread_data *map[REAL_MAX_JOBS];
 		struct timeval this_start;
 		int this_jobs = 0, left;
 
@@ -1636,6 +1695,7 @@ int main(int argc, char *argv[])
 	long ps;
 
 	sinit();
+	init_rand(&__fio_rand_state);
 
 	/*
 	 * We need locale for number printing, if it isn't set then just
@@ -1670,11 +1730,14 @@ int main(int argc, char *argv[])
 	}
 
 	startup_mutex = fio_mutex_init(0);
+	if (startup_mutex == NULL)
+		return 1;
 	writeout_mutex = fio_mutex_init(1);
+	if (writeout_mutex == NULL)
+		return 1;
 
 	set_genesis_time();
-
-	status_timer_arm();
+	create_disk_util_thread();
 
 	cgroup_list = smalloc(sizeof(*cgroup_list));
 	INIT_FLIST_HEAD(cgroup_list);
@@ -1692,8 +1755,7 @@ int main(int argc, char *argv[])
 
 	cgroup_kill(cgroup_list);
 	sfree(cgroup_list);
-	if (cgroup_mnt)
-		sfree(cgroup_mnt);
+	sfree(cgroup_mnt);
 
 	fio_mutex_remove(startup_mutex);
 	fio_mutex_remove(writeout_mutex);
