@@ -15,40 +15,25 @@
 
 void fio_mutex_remove(struct fio_mutex *mutex)
 {
-	close(mutex->mutex_fd);
 	munmap((void *) mutex, sizeof(*mutex));
 }
 
 struct fio_mutex *fio_mutex_init(int value)
 {
-	char mutex_name[] = "/tmp/.fio_mutex.XXXXXX";
 	struct fio_mutex *mutex = NULL;
 	pthread_mutexattr_t attr;
 	pthread_condattr_t cond;
-	int fd, ret, mflag;
-
-	fd = mkstemp(mutex_name);
-	if (fd < 0) {
-		perror("open mutex");
-		return NULL;
-	}
-
-	if (ftruncate(fd, sizeof(struct fio_mutex)) < 0) {
-		perror("ftruncate mutex");
-		goto err;
-	}
+	int ret, mflag;
 
 	mutex = (void *) mmap(NULL, sizeof(struct fio_mutex),
-				PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
+				PROT_READ | PROT_WRITE,
+				OS_MAP_ANON | MAP_SHARED, -1, 0);
 	if (mutex == MAP_FAILED) {
 		perror("mmap mutex");
-		close(fd);
 		mutex = NULL;
 		goto err;
 	}
 
-	unlink(mutex_name);
-	mutex->mutex_fd = fd;
 	mutex->value = value;
 
 	/*
@@ -65,14 +50,18 @@ struct fio_mutex *fio_mutex_init(int value)
 		log_err("pthread_mutexattr_init: %s\n", strerror(ret));
 		goto err;
 	}
+#ifdef FIO_HAVE_PSHARED_MUTEX
 	ret = pthread_mutexattr_setpshared(&attr, mflag);
 	if (ret) {
 		log_err("pthread_mutexattr_setpshared: %s\n", strerror(ret));
 		goto err;
 	}
+#endif
 
 	pthread_condattr_init(&cond);
+#ifdef FIO_HAVE_PSHARED_MUTEX
 	pthread_condattr_setpshared(&cond, mflag);
+#endif
 	pthread_cond_init(&mutex->cond, &cond);
 
 	ret = pthread_mutex_init(&mutex->lock, &attr);
@@ -81,12 +70,14 @@ struct fio_mutex *fio_mutex_init(int value)
 		goto err;
 	}
 
+	pthread_condattr_destroy(&cond);
+	pthread_mutexattr_destroy(&attr);
+
 	return mutex;
 err:
 	if (mutex)
 		fio_mutex_remove(mutex);
 
-	unlink(mutex_name);
 	return NULL;
 }
 
