@@ -6,10 +6,13 @@
 #include <string.h>
 #include <assert.h>
 #include <pthread.h>
+#include <libgen.h>
 
 #include "fio.h"
 #include "verify.h"
 #include "smalloc.h"
+#include "trim.h"
+#include "lib/rand.h"
 
 #include "crc/md5.h"
 #include "crc/crc64.h"
@@ -21,44 +24,53 @@
 #include "crc/sha512.h"
 #include "crc/sha1.h"
 
-static void fill_random_bytes(struct thread_data *td, void *p, unsigned int len)
-{
-	unsigned int todo;
-	int r;
+static void populate_hdr(struct thread_data *td, struct io_u *io_u,
+			 struct verify_header *hdr, unsigned int header_num,
+			 unsigned int header_len);
 
-	while (len) {
-		r = os_random_long(&td->verify_state);
-
-		/*
-		 * lrand48_r seems to be broken and only fill the bottom
-		 * 32-bits, even on 64-bit archs with 64-bit longs
-		 */
-		todo = sizeof(r);
-		if (todo > len)
-			todo = len;
-
-		memcpy(p, &r, todo);
-
-		len -= todo;
-		p += todo;
-	}
-}
-
-static void fill_pattern(struct thread_data *td, void *p, unsigned int len)
+void fill_pattern(struct thread_data *td, void *p, unsigned int len, struct io_u *io_u, unsigned long seed, int use_seed)
 {
 	switch (td->o.verify_pattern_bytes) {
 	case 0:
 		dprint(FD_VERIFY, "fill random bytes len=%u\n", len);
-		fill_random_bytes(td, p, len);
+		if (use_seed)
+			__fill_random_buf(p, len, seed);
+		else
+			io_u->rand_seed = fill_random_buf(p, len);
 		break;
 	case 1:
+		/*
+		 * See below write barrier comment
+		 */
+#if 0
+		read_barrier();
+		if (io_u->buf_filled_len >= len) {
+			dprint(FD_VERIFY, "using already filled verify pattern b=0 len=%u\n", len);
+			return;
+		}
+#endif
 		dprint(FD_VERIFY, "fill verify pattern b=0 len=%u\n", len);
 		memset(p, td->o.verify_pattern[0], len);
+		/*
+		 * We need to ensure that the pattern stores are seen before
+		 * the fill length store, or we could observe headers that
+		 * aren't valid to the extent notified by the fill length
+		 */
+		write_barrier();
+		io_u->buf_filled_len = len;
 		break;
 	default: {
 		unsigned int i = 0, size = 0;
 		unsigned char *b = p;
 
+#if 0
+		read_barrier();
+		if (io_u->buf_filled_len >= len) {
+			dprint(FD_VERIFY, "using already filled verify pattern b=%d len=%u\n",
+					td->o.verify_pattern_bytes, len);
+			return;
+		}
+#endif
 		dprint(FD_VERIFY, "fill verify pattern b=%d len=%u\n",
 					td->o.verify_pattern_bytes, len);
 
@@ -69,11 +81,42 @@ static void fill_pattern(struct thread_data *td, void *p, unsigned int len)
 			memcpy(b+i, td->o.verify_pattern, size);
 			i += size;
 		}
+		write_barrier();
+		io_u->buf_filled_len = len;
 		break;
 		}
 	}
 }
 
+static unsigned int get_hdr_inc(struct thread_data *td, struct io_u *io_u)
+{
+	unsigned int hdr_inc;
+
+	hdr_inc = io_u->buflen;
+	if (td->o.verify_interval)
+		hdr_inc = td->o.verify_interval;
+
+	return hdr_inc;
+}
+
+static void fill_pattern_headers(struct thread_data *td, struct io_u *io_u,
+				 unsigned long seed, int use_seed)
+{
+	unsigned int hdr_inc, header_num;
+	struct verify_header *hdr;
+	void *p = io_u->buf;
+
+	fill_pattern(td, p, io_u->buflen, io_u, seed, use_seed);
+
+	hdr_inc = get_hdr_inc(td, io_u);
+	header_num = 0;
+	for (; p < io_u->buf + io_u->buflen; p += hdr_inc) {
+		hdr = p;
+		populate_hdr(td, io_u, hdr, header_num, hdr_inc);
+		header_num++;
+	}
+}
+
 static void memswp(void *buf1, void *buf2, unsigned int len)
 {
 	char swap[200];
@@ -91,8 +134,8 @@ static void hexdump(void *buffer, int len)
 	int i;
 
 	for (i = 0; i < len; i++)
-		log_info("%02x", p[i]);
-	log_info("\n");
+		log_err("%02x", p[i]);
+	log_err("\n");
 }
 
 /*
@@ -100,7 +143,7 @@ static void hexdump(void *buffer, int len)
  */
 static inline unsigned int __hdr_size(int verify_type)
 {
-	unsigned int len = len;
+	unsigned int len = 0;
 
 	switch (verify_type) {
 	case VERIFY_NONE:
@@ -136,6 +179,9 @@ static inline unsigned int __hdr_size(int verify_type)
 	case VERIFY_SHA1:
 		len = sizeof(struct vhdr_sha1);
 		break;
+	case VERIFY_PATTERN:
+		len = 0;
+		break;
 	default:
 		log_err("fio: unknown verify header!\n");
 		assert(0);
@@ -156,243 +202,122 @@ static void *hdr_priv(struct verify_header *hdr)
 	return priv + sizeof(struct verify_header);
 }
 
+/*
+ * Verify container, pass info to verify handlers and allow them to
+ * pass info back in case of error
+ */
+struct vcont {
+	/*
+	 * Input
+	 */
+	struct io_u *io_u;
+	unsigned int hdr_num;
+	struct thread_data *td;
+
+	/*
+	 * Output, only valid in case of error
+	 */
+	const char *name;
+	void *good_crc;
+	void *bad_crc;
+	unsigned int crc_len;
+};
+
+static void dump_buf(char *buf, unsigned int len, unsigned long long offset,
+		     const char *type, struct fio_file *f)
+{
+	char *ptr, fname[256];
+	int ret, fd;
+
+	ptr = strdup(f->file_name);
+	strcpy(fname, basename(ptr));
+
+	sprintf(fname + strlen(fname), ".%llu.%s", offset, type);
+
+	fd = open(fname, O_CREAT | O_TRUNC | O_WRONLY, 0644);
+	if (fd < 0) {
+		perror("open verify buf file");
+		return;
+	}
+
+	while (len) {
+		ret = write(fd, buf, len);
+		if (!ret)
+			break;
+		else if (ret < 0) {
+			perror("write verify buf file");
+			break;
+		}
+		len -= ret;
+		buf += ret;
+	}
+
+	close(fd);
+	log_err("       %s data dumped as %s\n", type, fname);
+	free(ptr);
+}
+
+/*
+ * Dump the contents of the read block and re-generate the correct data
+ * and dump that too.
+ */
+static void dump_verify_buffers(struct verify_header *hdr, struct vcont *vc)
+{
+	struct thread_data *td = vc->td;
+	struct io_u *io_u = vc->io_u;
+	unsigned long hdr_offset;
+	struct io_u dummy;
+	void *buf;
+
+	/*
+	 * Dump the contents we just read off disk
+	 */
+	hdr_offset = vc->hdr_num * hdr->len;
+
+	dump_buf(io_u->buf + hdr_offset, hdr->len, io_u->offset + hdr_offset,
+			"received", vc->io_u->file);
+
+	/*
+	 * Allocate a new buf and re-generate the original data
+	 */
+	buf = malloc(io_u->buflen);
+	dummy = *io_u;
+	dummy.buf = buf;
+	dummy.rand_seed = hdr->rand_seed;
+	dummy.buf_filled_len = 0;
+
+	fill_pattern_headers(td, &dummy, hdr->rand_seed, 1);
+
+	dump_buf(buf + hdr_offset, hdr->len, io_u->offset + hdr_offset,
+			"expected", vc->io_u->file);
+	free(buf);
+}
+
+static void log_verify_failure(struct verify_header *hdr, struct vcont *vc)
+{
+	unsigned long long offset;
+
+	offset = vc->io_u->offset;
+	offset += vc->hdr_num * hdr->len;
+	log_err("%.8s: verify failed at file %s offset %llu, length %u\n",
+			vc->name, vc->io_u->file->file_name, offset, hdr->len);
+
+	if (vc->good_crc && vc->bad_crc) {
+		log_err("       Expected CRC: ");
+		hexdump(vc->good_crc, vc->crc_len);
+		log_err("       Received CRC: ");
+		hexdump(vc->bad_crc, vc->crc_len);
+	}
+
+	dump_verify_buffers(hdr, vc);
+}
+
 /*
  * Return data area 'header_num'
  */
-static inline void *io_u_verify_off(struct verify_header *hdr,
-				    struct io_u *io_u, unsigned char header_num)
+static inline void *io_u_verify_off(struct verify_header *hdr, struct vcont *vc)
 {
-	return io_u->buf + header_num * hdr->len + hdr_size(hdr);
-}
-
-static int verify_io_u_meta(struct verify_header *hdr, struct thread_data *td,
-			    struct io_u *io_u, unsigned int header_num)
-{
-	struct vhdr_meta *vh = hdr_priv(hdr);
-
-	dprint(FD_VERIFY, "meta verify io_u %p, len %u\n", io_u, hdr->len);
-
-	if (vh->offset != io_u->offset + header_num * td->o.verify_interval) {
-		log_err("meta: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_sha512(struct verify_header *hdr, struct io_u *io_u,
-			      unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_sha512 *vh = hdr_priv(hdr);
-	uint8_t sha512[128];
-	struct sha512_ctx sha512_ctx = {
-		.buf = sha512,
-	};
-
-	dprint(FD_VERIFY, "sha512 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	sha512_init(&sha512_ctx);
-	sha512_update(&sha512_ctx, p, hdr->len - hdr_size(hdr));
-
-	if (memcmp(vh->sha512, sha512_ctx.buf, sizeof(sha512))) {
-		log_err("sha512: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->sha512, sizeof(vh->sha512));
-		hexdump(sha512_ctx.buf, sizeof(sha512));
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_sha256(struct verify_header *hdr, struct io_u *io_u,
-			      unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_sha256 *vh = hdr_priv(hdr);
-	uint8_t sha256[64];
-	struct sha256_ctx sha256_ctx = {
-		.buf = sha256,
-	};
-
-	dprint(FD_VERIFY, "sha256 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	sha256_init(&sha256_ctx);
-	sha256_update(&sha256_ctx, p, hdr->len - hdr_size(hdr));
-
-	if (memcmp(vh->sha256, sha256_ctx.buf, sizeof(sha256))) {
-		log_err("sha256: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->sha256, sizeof(vh->sha256));
-		hexdump(sha256_ctx.buf, sizeof(sha256));
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_sha1(struct verify_header *hdr, struct io_u *io_u,
-			    unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_sha1 *vh = hdr_priv(hdr);
-	uint32_t sha1[5];
-	struct sha1_ctx sha1_ctx = {
-		.H = sha1,
-	};
-
-	dprint(FD_VERIFY, "sha1 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	sha1_init(&sha1_ctx);
-	sha1_update(&sha1_ctx, p, hdr->len - hdr_size(hdr));
-
-	if (memcmp(vh->sha1, sha1_ctx.H, sizeof(sha1))) {
-		log_err("sha1: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->sha1, sizeof(vh->sha1));
-		hexdump(sha1_ctx.H, sizeof(sha1));
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_crc7(struct verify_header *hdr, struct io_u *io_u,
-			    unsigned char header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_crc7 *vh = hdr_priv(hdr);
-	unsigned char c;
-
-	dprint(FD_VERIFY, "crc7 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	c = crc7(p, hdr->len - hdr_size(hdr));
-
-	if (c != vh->crc7) {
-		log_err("crc7: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc7: wanted %x, got %x\n", vh->crc7, c);
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_crc16(struct verify_header *hdr, struct io_u *io_u,
-			     unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_crc16 *vh = hdr_priv(hdr);
-	unsigned short c;
-
-	dprint(FD_VERIFY, "crc16 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	c = crc16(p, hdr->len - hdr_size(hdr));
-
-	if (c != vh->crc16) {
-		log_err("crc16: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc16: wanted %x, got %x\n", vh->crc16, c);
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_crc64(struct verify_header *hdr, struct io_u *io_u,
-			     unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_crc64 *vh = hdr_priv(hdr);
-	unsigned long long c;
-
-	dprint(FD_VERIFY, "crc64 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	c = crc64(p, hdr->len - hdr_size(hdr));
-
-	if (c != vh->crc64) {
-		log_err("crc64: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len,
-				hdr->len);
-		log_err("crc64: wanted %llx, got %llx\n",
-					(unsigned long long) vh->crc64, c);
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_crc32(struct verify_header *hdr, struct io_u *io_u,
-			     unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_crc32 *vh = hdr_priv(hdr);
-	uint32_t c;
-
-	dprint(FD_VERIFY, "crc32 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	c = crc32(p, hdr->len - hdr_size(hdr));
-
-	if (c != vh->crc32) {
-		log_err("crc32: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc32: wanted %x, got %x\n", vh->crc32, c);
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_crc32c(struct verify_header *hdr, struct io_u *io_u,
-			      unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_crc32 *vh = hdr_priv(hdr);
-	uint32_t c;
-
-	dprint(FD_VERIFY, "crc32c verify io_u %p, len %u\n", io_u, hdr->len);
-
-	if (hdr->verify_type == VERIFY_CRC32C_INTEL)
-		c = crc32c_intel(p, hdr->len - hdr_size(hdr));
-	else
-		c = crc32c(p, hdr->len - hdr_size(hdr));
-
-	if (c != vh->crc32) {
-		log_err("crc32c: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc32c: wanted %x, got %x\n", vh->crc32, c);
-		return EILSEQ;
-	}
-
-	return 0;
-}
-
-static int verify_io_u_md5(struct verify_header *hdr, struct io_u *io_u,
-			   unsigned int header_num)
-{
-	void *p = io_u_verify_off(hdr, io_u, header_num);
-	struct vhdr_md5 *vh = hdr_priv(hdr);
-	uint32_t hash[MD5_HASH_WORDS];
-	struct md5_ctx md5_ctx = {
-		.hash = hash,
-	};
-
-	dprint(FD_VERIFY, "md5 verify io_u %p, len %u\n", io_u, hdr->len);
-
-	md5_init(&md5_ctx);
-	md5_update(&md5_ctx, p, hdr->len - hdr_size(hdr));
-
-	if (memcmp(vh->md5_digest, md5_ctx.hash, sizeof(hash))) {
-		log_err("md5: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->md5_digest, sizeof(vh->md5_digest));
-		hexdump(md5_ctx.hash, sizeof(hash));
-		return EILSEQ;
-	}
-
-	return 0;
+	return vc->io_u->buf + vc->hdr_num * hdr->len + hdr_size(hdr);
 }
 
 static unsigned int hweight8(unsigned int w)
@@ -403,10 +328,18 @@ static unsigned int hweight8(unsigned int w)
 	return (res + (res >> 4)) & 0x0F;
 }
 
-int verify_io_u_pattern(char *pattern, unsigned long pattern_size,
-			char *buf, unsigned int len, unsigned int mod)
+static int verify_io_u_pattern(struct verify_header *hdr, struct vcont *vc)
 {
-	unsigned int i;
+	struct thread_data *td = vc->td;
+	struct io_u *io_u = vc->io_u;
+	char *buf, *pattern;
+	unsigned int header_size = __hdr_size(td->o.verify);
+	unsigned int len, mod, i;
+
+	pattern = td->o.verify_pattern;
+	buf = (void *) hdr + header_size;
+	len = get_hdr_inc(td, io_u) - header_size;
+	mod = header_size % td->o.verify_pattern_bytes;
 
 	for (i = 0; i < len; i++) {
 		if (buf[i] != pattern[mod]) {
@@ -416,16 +349,248 @@ int verify_io_u_pattern(char *pattern, unsigned long pattern_size,
 			log_err("fio: got pattern %x, wanted %x. Bad bits %d\n",
 				buf[i], pattern[mod], bits);
 			log_err("fio: bad pattern block offset %u\n", i);
+			dump_verify_buffers(hdr, vc);
 			return EILSEQ;
 		}
 		mod++;
-		if (mod == pattern_size)
+		if (mod == td->o.verify_pattern_bytes)
 			mod = 0;
 	}
 
 	return 0;
 }
 
+static int verify_io_u_meta(struct verify_header *hdr, struct vcont *vc)
+{
+	struct thread_data *td = vc->td;
+	struct vhdr_meta *vh = hdr_priv(hdr);
+	struct io_u *io_u = vc->io_u;
+	int ret = EILSEQ;
+
+	dprint(FD_VERIFY, "meta verify io_u %p, len %u\n", io_u, hdr->len);
+
+	if (vh->offset == io_u->offset + vc->hdr_num * td->o.verify_interval)
+		ret = 0;
+
+	if (td->o.verify_pattern_bytes)
+		ret |= verify_io_u_pattern(hdr, vc);
+
+	if (!ret)
+		return 0;
+
+	vc->name = "meta";
+	log_verify_failure(hdr, vc);
+	return ret;
+}
+
+static int verify_io_u_sha512(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_sha512 *vh = hdr_priv(hdr);
+	uint8_t sha512[128];
+	struct sha512_ctx sha512_ctx = {
+		.buf = sha512,
+	};
+
+	dprint(FD_VERIFY, "sha512 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	sha512_init(&sha512_ctx);
+	sha512_update(&sha512_ctx, p, hdr->len - hdr_size(hdr));
+
+	if (!memcmp(vh->sha512, sha512_ctx.buf, sizeof(sha512)))
+		return 0;
+
+	vc->name = "sha512";
+	vc->good_crc = vh->sha512;
+	vc->bad_crc = sha512_ctx.buf;
+	vc->crc_len = sizeof(vh->sha512);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_sha256(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_sha256 *vh = hdr_priv(hdr);
+	uint8_t sha256[64];
+	struct sha256_ctx sha256_ctx = {
+		.buf = sha256,
+	};
+
+	dprint(FD_VERIFY, "sha256 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	sha256_init(&sha256_ctx);
+	sha256_update(&sha256_ctx, p, hdr->len - hdr_size(hdr));
+
+	if (!memcmp(vh->sha256, sha256_ctx.buf, sizeof(sha256)))
+		return 0;
+
+	vc->name = "sha256";
+	vc->good_crc = vh->sha256;
+	vc->bad_crc = sha256_ctx.buf;
+	vc->crc_len = sizeof(vh->sha256);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_sha1(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_sha1 *vh = hdr_priv(hdr);
+	uint32_t sha1[5];
+	struct sha1_ctx sha1_ctx = {
+		.H = sha1,
+	};
+
+	dprint(FD_VERIFY, "sha1 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	sha1_init(&sha1_ctx);
+	sha1_update(&sha1_ctx, p, hdr->len - hdr_size(hdr));
+
+	if (!memcmp(vh->sha1, sha1_ctx.H, sizeof(sha1)))
+		return 0;
+
+	vc->name = "sha1";
+	vc->good_crc = vh->sha1;
+	vc->bad_crc = sha1_ctx.H;
+	vc->crc_len = sizeof(vh->sha1);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_crc7(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_crc7 *vh = hdr_priv(hdr);
+	unsigned char c;
+
+	dprint(FD_VERIFY, "crc7 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	c = crc7(p, hdr->len - hdr_size(hdr));
+
+	if (c == vh->crc7)
+		return 0;
+
+	vc->name = "crc7";
+	vc->good_crc = &vh->crc7;
+	vc->bad_crc = &c;
+	vc->crc_len = 1;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_crc16(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_crc16 *vh = hdr_priv(hdr);
+	unsigned short c;
+
+	dprint(FD_VERIFY, "crc16 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	c = crc16(p, hdr->len - hdr_size(hdr));
+
+	if (c == vh->crc16)
+		return 0;
+
+	vc->name = "crc16";
+	vc->good_crc = &vh->crc16;
+	vc->bad_crc = &c;
+	vc->crc_len = 2;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_crc64(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_crc64 *vh = hdr_priv(hdr);
+	unsigned long long c;
+
+	dprint(FD_VERIFY, "crc64 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	c = crc64(p, hdr->len - hdr_size(hdr));
+
+	if (c == vh->crc64)
+		return 0;
+
+	vc->name = "crc64";
+	vc->good_crc = &vh->crc64;
+	vc->bad_crc = &c;
+	vc->crc_len = 8;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_crc32(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_crc32 *vh = hdr_priv(hdr);
+	uint32_t c;
+
+	dprint(FD_VERIFY, "crc32 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	c = crc32(p, hdr->len - hdr_size(hdr));
+
+	if (c == vh->crc32)
+		return 0;
+
+	vc->name = "crc32";
+	vc->good_crc = &vh->crc32;
+	vc->bad_crc = &c;
+	vc->crc_len = 4;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_crc32c(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_crc32 *vh = hdr_priv(hdr);
+	uint32_t c;
+
+	dprint(FD_VERIFY, "crc32c verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	if (hdr->verify_type == VERIFY_CRC32C_INTEL)
+		c = crc32c_intel(p, hdr->len - hdr_size(hdr));
+	else
+		c = crc32c(p, hdr->len - hdr_size(hdr));
+
+	if (c == vh->crc32)
+		return 0;
+
+	vc->name = "crc32c";
+	vc->good_crc = &vh->crc32;
+	vc->bad_crc = &c;
+	vc->crc_len = 4;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
+static int verify_io_u_md5(struct verify_header *hdr, struct vcont *vc)
+{
+	void *p = io_u_verify_off(hdr, vc);
+	struct vhdr_md5 *vh = hdr_priv(hdr);
+	uint32_t hash[MD5_HASH_WORDS];
+	struct md5_ctx md5_ctx = {
+		.hash = hash,
+	};
+
+	dprint(FD_VERIFY, "md5 verify io_u %p, len %u\n", vc->io_u, hdr->len);
+
+	md5_init(&md5_ctx);
+	md5_update(&md5_ctx, p, hdr->len - hdr_size(hdr));
+
+	if (!memcmp(vh->md5_digest, md5_ctx.hash, sizeof(hash)))
+		return 0;
+
+	vc->name = "md5";
+	vc->good_crc = vh->md5_digest;
+	vc->bad_crc = md5_ctx.hash;
+	vc->crc_len = sizeof(hash);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
+}
+
 /*
  * Push IO verification to a separate thread
  */
@@ -451,91 +616,115 @@ int verify_io_u_async(struct thread_data *td, struct io_u *io_u)
 	return 0;
 }
 
+static int verify_trimmed_io_u(struct thread_data *td, struct io_u *io_u)
+{
+	static char zero_buf[1024];
+	unsigned int this_len, len;
+	int ret = 0;
+	void *p;
+
+	if (!td->o.trim_zero)
+		return 0;
+
+	len = io_u->buflen;
+	p = io_u->buf;
+	do {
+		this_len = sizeof(zero_buf);
+		if (this_len > len)
+			this_len = len;
+		if (memcmp(p, zero_buf, this_len)) {
+			ret = EILSEQ;
+			break;
+		}
+		len -= this_len;
+		p += this_len;
+	} while (len);
+
+	if (!ret)
+		return 0;
+
+	log_err("trim: verify failed at file %s offset %llu, length %lu"
+		", block offset %lu\n",
+			io_u->file->file_name, io_u->offset, io_u->buflen,
+			(unsigned long) (p - io_u->buf));
+	return ret;
+}
+
 int verify_io_u(struct thread_data *td, struct io_u *io_u)
 {
 	struct verify_header *hdr;
-	unsigned int hdr_size, hdr_inc, hdr_num = 0;
+	unsigned int header_size, hdr_inc, hdr_num = 0;
 	void *p;
 	int ret;
 
 	if (td->o.verify == VERIFY_NULL || io_u->ddir != DDIR_READ)
 		return 0;
+	if (io_u->flags & IO_U_F_TRIMMED) {
+		ret = verify_trimmed_io_u(td, io_u);
+		goto done;
+	}
 
-	hdr_inc = io_u->buflen;
-	if (td->o.verify_interval)
-		hdr_inc = td->o.verify_interval;
+	hdr_inc = get_hdr_inc(td, io_u);
 
 	ret = 0;
 	for (p = io_u->buf; p < io_u->buf + io_u->buflen;
 	     p += hdr_inc, hdr_num++) {
-		if (ret && td->o.verify_fatal) {
-			td->terminate = 1;
+		struct vcont vc = {
+			.io_u		= io_u,
+			.hdr_num	= hdr_num,
+			.td		= td,
+		};
+
+		if (ret && td->o.verify_fatal)
 			break;
-		}
-		hdr_size = __hdr_size(td->o.verify);
+
+		header_size = __hdr_size(td->o.verify);
 		if (td->o.verify_offset)
-			memswp(p, p + td->o.verify_offset, hdr_size);
+			memswp(p, p + td->o.verify_offset, header_size);
 		hdr = p;
 
 		if (hdr->fio_magic != FIO_HDR_MAGIC) {
-			log_err("Bad verify header %x at %llu\n",
-					hdr->fio_magic,
-					io_u->offset + hdr_num * hdr->len);
+			log_err("verify: bad magic header %x, wanted %x at file %s offset %llu, length %u\n",
+				hdr->fio_magic, FIO_HDR_MAGIC,
+				io_u->file->file_name,
+				io_u->offset + hdr_num * hdr->len, hdr->len);
 			return EILSEQ;
 		}
 
-		if (td->o.verify_pattern_bytes) {
-			dprint(FD_VERIFY, "pattern verify io_u %p, len %u\n",
-								io_u, hdr->len);
-			ret = verify_io_u_pattern(td->o.verify_pattern,
-				  td->o.verify_pattern_bytes,
-				  p + hdr_size,
-				  hdr_inc - hdr_size,
-				  hdr_size % td->o.verify_pattern_bytes);
-			/*
-			 * Also verify the meta data, if applicable
-			 */
-			if (hdr->verify_type == VERIFY_META)
-				ret |= verify_io_u_meta(hdr, td, io_u, hdr_num);
-
-			if (ret)
-				log_err("fio: verify failed at %llu/%u\n",
-					io_u->offset + hdr_num * hdr->len,
-					hdr->len);
-			continue;
-		}
-
 		switch (hdr->verify_type) {
 		case VERIFY_MD5:
-			ret = verify_io_u_md5(hdr, io_u, hdr_num);
+			ret = verify_io_u_md5(hdr, &vc);
 			break;
 		case VERIFY_CRC64:
-			ret = verify_io_u_crc64(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc64(hdr, &vc);
 			break;
 		case VERIFY_CRC32C:
 		case VERIFY_CRC32C_INTEL:
-			ret = verify_io_u_crc32c(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc32c(hdr, &vc);
 			break;
 		case VERIFY_CRC32:
-			ret = verify_io_u_crc32(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc32(hdr, &vc);
 			break;
 		case VERIFY_CRC16:
-			ret = verify_io_u_crc16(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc16(hdr, &vc);
 			break;
 		case VERIFY_CRC7:
-			ret = verify_io_u_crc7(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc7(hdr, &vc);
 			break;
 		case VERIFY_SHA256:
-			ret = verify_io_u_sha256(hdr, io_u, hdr_num);
+			ret = verify_io_u_sha256(hdr, &vc);
 			break;
 		case VERIFY_SHA512:
-			ret = verify_io_u_sha512(hdr, io_u, hdr_num);
+			ret = verify_io_u_sha512(hdr, &vc);
 			break;
 		case VERIFY_META:
-			ret = verify_io_u_meta(hdr, td, io_u, hdr_num);
+			ret = verify_io_u_meta(hdr, &vc);
 			break;
 		case VERIFY_SHA1:
-			ret = verify_io_u_sha1(hdr, io_u, hdr_num);
+			ret = verify_io_u_sha1(hdr, &vc);
+			break;
+		case VERIFY_PATTERN:
+			ret = verify_io_u_pattern(hdr, &vc);
 			break;
 		default:
 			log_err("Bad verify type %u\n", hdr->verify_type);
@@ -543,6 +732,10 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 		}
 	}
 
+done:
+	if (ret && td->o.verify_fatal)
+		td->terminate = 1;
+
 	return ret;
 }
 
@@ -643,94 +836,95 @@ static void fill_md5(struct verify_header *hdr, void *p, unsigned int len)
 	md5_update(&md5_ctx, p, len);
 }
 
+static void populate_hdr(struct thread_data *td, struct io_u *io_u,
+			 struct verify_header *hdr, unsigned int header_num,
+			 unsigned int header_len)
+{
+	unsigned int data_len;
+	void *data, *p;
+
+	p = (void *) hdr;
+
+	hdr->fio_magic = FIO_HDR_MAGIC;
+	hdr->len = header_len;
+	hdr->verify_type = td->o.verify;
+	hdr->rand_seed = io_u->rand_seed;
+	data_len = header_len - hdr_size(hdr);
+
+	data = p + hdr_size(hdr);
+	switch (td->o.verify) {
+	case VERIFY_MD5:
+		dprint(FD_VERIFY, "fill md5 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_md5(hdr, data, data_len);
+		break;
+	case VERIFY_CRC64:
+		dprint(FD_VERIFY, "fill crc64 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_crc64(hdr, data, data_len);
+		break;
+	case VERIFY_CRC32C:
+	case VERIFY_CRC32C_INTEL:
+		dprint(FD_VERIFY, "fill crc32c io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_crc32c(hdr, data, data_len);
+		break;
+	case VERIFY_CRC32:
+		dprint(FD_VERIFY, "fill crc32 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_crc32(hdr, data, data_len);
+		break;
+	case VERIFY_CRC16:
+		dprint(FD_VERIFY, "fill crc16 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_crc16(hdr, data, data_len);
+		break;
+	case VERIFY_CRC7:
+		dprint(FD_VERIFY, "fill crc7 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_crc7(hdr, data, data_len);
+		break;
+	case VERIFY_SHA256:
+		dprint(FD_VERIFY, "fill sha256 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_sha256(hdr, data, data_len);
+		break;
+	case VERIFY_SHA512:
+		dprint(FD_VERIFY, "fill sha512 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_sha512(hdr, data, data_len);
+		break;
+	case VERIFY_META:
+		dprint(FD_VERIFY, "fill meta io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_meta(hdr, td, io_u, header_num);
+		break;
+	case VERIFY_SHA1:
+		dprint(FD_VERIFY, "fill sha1 io_u %p, len %u\n",
+						io_u, hdr->len);
+		fill_sha1(hdr, data, data_len);
+		break;
+	case VERIFY_PATTERN:
+		/* nothing to do here */
+		break;
+	default:
+		log_err("fio: bad verify type: %d\n", td->o.verify);
+		assert(0);
+	}
+	if (td->o.verify_offset)
+		memswp(p, p + td->o.verify_offset, hdr_size(hdr));
+}
+
 /*
  * fill body of io_u->buf with random data and add a header with the
- * crc32 or md5 sum of that data.
+ * checksum of choice
  */
 void populate_verify_io_u(struct thread_data *td, struct io_u *io_u)
 {
-	struct verify_header *hdr;
-	void *p = io_u->buf, *data;
-	unsigned int hdr_inc, data_len, header_num = 0;
-
 	if (td->o.verify == VERIFY_NULL)
 		return;
 
-	fill_pattern(td, p, io_u->buflen);
-
-	hdr_inc = io_u->buflen;
-	if (td->o.verify_interval)
-		hdr_inc = td->o.verify_interval;
-
-	for (; p < io_u->buf + io_u->buflen; p += hdr_inc) {
-		hdr = p;
-
-		hdr->fio_magic = FIO_HDR_MAGIC;
-		hdr->verify_type = td->o.verify;
-		hdr->len = hdr_inc;
-		data_len = hdr_inc - hdr_size(hdr);
-
-		data = p + hdr_size(hdr);
-		switch (td->o.verify) {
-		case VERIFY_MD5:
-			dprint(FD_VERIFY, "fill md5 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_md5(hdr, data, data_len);
-			break;
-		case VERIFY_CRC64:
-			dprint(FD_VERIFY, "fill crc64 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_crc64(hdr, data, data_len);
-			break;
-		case VERIFY_CRC32C:
-		case VERIFY_CRC32C_INTEL:
-			dprint(FD_VERIFY, "fill crc32c io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_crc32c(hdr, data, data_len);
-			break;
-		case VERIFY_CRC32:
-			dprint(FD_VERIFY, "fill crc32 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_crc32(hdr, data, data_len);
-			break;
-		case VERIFY_CRC16:
-			dprint(FD_VERIFY, "fill crc16 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_crc16(hdr, data, data_len);
-			break;
-		case VERIFY_CRC7:
-			dprint(FD_VERIFY, "fill crc7 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_crc7(hdr, data, data_len);
-			break;
-		case VERIFY_SHA256:
-			dprint(FD_VERIFY, "fill sha256 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_sha256(hdr, data, data_len);
-			break;
-		case VERIFY_SHA512:
-			dprint(FD_VERIFY, "fill sha512 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_sha512(hdr, data, data_len);
-			break;
-		case VERIFY_META:
-			dprint(FD_VERIFY, "fill meta io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_meta(hdr, td, io_u, header_num);
-			break;
-		case VERIFY_SHA1:
-			dprint(FD_VERIFY, "fill sha1 io_u %p, len %u\n",
-							io_u, hdr->len);
-			fill_sha1(hdr, data, data_len);
-			break;
-		default:
-			log_err("fio: bad verify type: %d\n", td->o.verify);
-			assert(0);
-		}
-		if (td->o.verify_offset)
-			memswp(p, p + td->o.verify_offset, hdr_size(hdr));
-		header_num++;
-	}
+	fill_pattern_headers(td, io_u, 0, 0);
 }
 
 int get_next_verify(struct thread_data *td, struct io_u *io_u)
@@ -748,16 +942,25 @@ int get_next_verify(struct thread_data *td, struct io_u *io_u)
 
 		ipo = rb_entry(n, struct io_piece, rb_node);
 		rb_erase(n, &td->io_hist_tree);
+		assert(ipo->flags & IP_F_ONRB);
+		ipo->flags &= ~IP_F_ONRB;
 	} else if (!flist_empty(&td->io_hist_list)) {
 		ipo = flist_entry(td->io_hist_list.next, struct io_piece, list);
 		flist_del(&ipo->list);
+		assert(ipo->flags & IP_F_ONLIST);
+		ipo->flags &= ~IP_F_ONLIST;
 	}
 
 	if (ipo) {
+		td->io_hist_len--;
+
 		io_u->offset = ipo->offset;
 		io_u->buflen = ipo->len;
 		io_u->file = ipo->file;
 
+		if (ipo->flags & IP_F_TRIMMED)
+			io_u->flags |= IO_U_F_TRIMMED;
+
 		if (!fio_file_open(io_u->file)) {
 			int r = td_io_open_file(td, io_u->file);
 
@@ -773,6 +976,8 @@ int get_next_verify(struct thread_data *td, struct io_u *io_u)
 		io_u->ddir = DDIR_READ;
 		io_u->xfer_buf = io_u->buf;
 		io_u->xfer_buflen = io_u->buflen;
+
+		remove_trim_entry(td, ipo);
 		free(ipo);
 		dprint(FD_VERIFY, "get_next_verify: ret io_u %p\n", io_u);
 		return 0;
@@ -838,7 +1043,8 @@ static void *verify_async_thread(void *data)
 
 	if (ret) {
 		td_verror(td, ret, "async_verify");
-		td->terminate = 1;
+		if (td->o.verify_fatal)
+			td->terminate = 1;
 	}
 
 done:
@@ -853,12 +1059,16 @@ done:
 int verify_async_init(struct thread_data *td)
 {
 	int i, ret;
+	pthread_attr_t attr;
+
+	pthread_attr_init(&attr);
+	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN);
 
 	td->verify_thread_exit = 0;
 
 	td->verify_threads = malloc(sizeof(pthread_t) * td->o.verify_async);
 	for (i = 0; i < td->o.verify_async; i++) {
-		ret = pthread_create(&td->verify_threads[i], NULL,
+		ret = pthread_create(&td->verify_threads[i], &attr,
 					verify_async_thread, td);
 		if (ret) {
 			log_err("fio: async verify creation failed: %s\n",
@@ -874,6 +1084,8 @@ int verify_async_init(struct thread_data *td)
 		td->nr_verify_threads++;
 	}
 
+	pthread_attr_destroy(&attr);
+
 	if (i != td->o.verify_async) {
 		log_err("fio: only %d verify threads started, exiting\n", i);
 		td->verify_thread_exit = 1;
